#!/usr/bin/perl

#use lib "/home/j/xuen/tools/lib/perl";
use TBNode;

#handling command-line parameters
use Getopt::Std ;

##################################################
# Optional arguments:                            #
# -t: trace mode                                 #
# -p: pbfile name specified                      #
# Obligatory arguments:                          #
# Input phrase structure treebank file directory #
# output dependency treebank file directory      #
###################################################

my %options;
$ok = getopts('tp:', \%options);
($ok) || die("Illegal optional arguments\n");


my $keep_trace = 0;

if(defined $options{"t"}){
    $keep_trace = 1;
}

undef $pbfile;
undef %pb;  #store pb annotation of a sentence
if(defined $options{"p"}){
    $pbfile = $options{"p"};
    #&readPB($pbfile); #populate pb, this is optional
    print $pbfile, "\n";
}

my $outputdir = $ARGV[0];
$outputdir ne "" || die "Empty output directory\n";
if(-e "$outputdir"){
    print "Find output in <$outputdir>\n";
}else{
    print "Creating <$outputdir> directory...";
    `mkdir $outputdir`;
}

undef %heads;

foreach my $index (1..$#ARGV){
    my $file = $ARGV[$index];
    $outfile = `basename $file`;
    $outfile =~ s/(chtb_\d+)\..*/$1\.dep/;
    print $outfile;
    open(OUT, ">$outputdir/$outfile");
    processfile("$file", \*OUT);
}

sub processfile{
    print STDERR "Processing file $file\n";
    my ($file, $outfh) = @_;
    my $basefile = `basename $file`;

    open(F, $file);
    #reset sentence count for each file
    $sentcount=0;  
    #reset the terminal map;
    undef %terminalMap; 

    my $fh = \*F;
    while (defined($tbn = TBNode::nextSentence($fh))) {
	#do stuff with a tbnode object $tbn ...
	if($tbn){
	    &termnoMap($tbn);
	    &annotate_heads($tbn);
	    #list of predicates for head of each arg
	    #assuming that the heads has already been mapped and stored in $heads
	    undef %argrel; #arguments for all predicates in a sentence 
	    undef %rels; #predicates for all arguments in a sentence
	    &parsePB($basefile, $sentcount); #populate %argrel and %rels
	    &collect_triples($outfh, $tbn, $sentcount);
	    print $outfh "\n";
	}
	$sentcount++;
    }
    close(F);
}

sub readPB{
    #read PB into a hash that is keyed by file number and sentence number
    my ($pbfile) = @_;
    open(PB, $pbfile);
    while(<PB>){
	my @parts = split;
	my $key = join(" ", @parts[0..1]);
	my $val = join(" ", @parts[2..$#parts]);
	#print "key = $key, val = $val\n";
	if (defined($pb{$key})){
	    $pb{$key} .= "|";
	    $pb{$key} .= $val;
	}else{
	    $pb{$key} = $val;
	}
    }
    close(PB);
}

sub parsePB{
    #parse the argument structure for each predicate, given a sentence
    my ($file, $sentcount) = @_;
    $file =~ s/\.head\.idt\s*$//;
    my $key = "$file $sentcount";
    my $pastr = $pb{$key};
    #print "$key pastr = $pastr\n";
    my @pas = split(/\|/, $pastr);

    foreach my $pa (@pas){
	#print "$pa\n";
	my @args = split(/\s+/,$pa); 
	my $relindex = $args[0];
	my $fid = $args[2];	
	#print "rel = $relindex $fid\n";
	#populate the predicate in each sentence
	if($rels{$sentcount}){ 
	    $rels{$sentcount} .= " ".$relindex."|".$fid;
	}else{
	    $rels{$sentcount} = $relindex."|".$fid;
	}

	for (my $i = 4; $i<=$#args; $i++){
	    my $arg = $args[$i];
	    if(!($arg =~ /rel$/i)){
		$arg =~ /([^\t ]+)\-(ARG[^\t ]+)/;
		my $addr = $1;
		my $lbl = $2;
		if($addr =~ /\*/){
		    #chained elements
		    my @addrs = split(/\*/, $addr);
		    my @sortedaddrs = sort(@addrs);

		    for (my $j=0; $j<=$#sortedaddrs; $j++){
			if($heads{$sortedaddrs[$j]}){
			    if($heads{$sortedaddrs[$j]} ne "TRACE"){
				$argrel{"$relindex|$heads{$sortedaddrs[$j]}"} = $lbl;
			    }else{
				#print "Node $key $sortedaddrs[$j] is a trace\n";
			    }
			}else{
			    print "No head for $key $sortedaddrs[$j]\n";
			}
		    }
		}elsif($addr=~/\,/){
		    #discontinuous elements
		    my @addrs = split(/\,/, $addr);
		    my @sortedaddrs = sort(@addrs);
		    if($heads{$sortedaddrs[0]}){
			if($heads{$sortedaddrs[0]} ne "TRACE"){
			    $argrel{"$relindex|$heads{$sortedaddrs[0]}"} = $lbl;
			}else{
			    #print "Node $key $sortedaddrs[0] is a trace\n";
			}
		    }else{
			print "No head found for $key $sortedaddrs[$0]\n";
		    }
		    for (my $j=1; $j<=$#sortedaddrs; $j++){
			if($heads{$sortedaddrs[$j]}){
			    if($heads{$sortedaddrs[$j]} ne "TRACE"){
				$argrel{"$relindex|$heads{$sortedaddrs[$j]}"} = "C-".$lbl;
			    }else{
				#print "Node $key $sortedaddrs[$j] is a trace\n";
			    }
			}else{
			    print "No head found for $key $sortedaddrs[$j]\n";
			}
		    }
		}else{
		    if($heads{$addr}){
			if($heads{$addr} ne "TRACE"){
			    $argrel{"$relindex|$heads{$addr}"} = $lbl;
			}else{
			    #print "Node $key $addr is a trace\n";
			}
		    }else{
			print "No head found for $key $addr\n";
		    }
		}
	    }
	}
    }
}
	


sub termnoMap{
    #a map between terminal numbers that take into account traces and 
    #terminal numbers that do not take traces into account
    my ($node) = @_;
    my $code = sub{ my $n = shift; return $n->height()==0;};
    my @terminals = $node->find($code);
    my $index = 1;
    foreach my $terminal (@terminals){
	if(defined($terminal)){
	    if(isTrace($terminal)){
		$terminalMap{$terminal->termno()} = "TRACE";
	    }else{
		$terminalMap{$terminal->termno()} = $index;
		$index++;
	    }
	}
    }
}



sub annotate_heads{
    #print "Annotating heads...\n";
    my ($node) = @_;
    my $code = sub{ my $n = shift; return $n->height()==0;};
    my @terminals = $node->find($code);
    foreach my $terminal (@terminals){

	my $value = $terminalMap{$terminal->termno()};
	my $key = createID($terminal);
	$heads{$key} = $value; #a terminal is always the head of itself
	my $p = $terminal;
	if(!isTrace($p)){
	    while(defined($p) && isHead($p)){  #head percolation
		my $key = createID($p); 
		$heads{$key} = $value; #all nodes that have this lexical head       
		$p=$p->parent();
	    }
	    
	    #is the head of parent as well
	    if(defined($p)){
		my $key = createID($p); 
		$heads{$key} = $value; #all nodes that have this lexical head
	    }  
	}else{
	    #when the terminal is a trace
	    my $pp = $p->parent(); #parent of terminal
	    if(defined($pp)){
		my $key = createID($pp);
		$heads{$key} = $value;
	    }
	}
    }
}

sub collect_triples{
    #print "Collecting triples...\n";
    my ($outfh, $node, $sentcount) = @_;

    #retrieve pa 
    my $relstr = $rels{$sentcount};
    #print "relstr = ", $relstr, "\n";
    my @preds = split(/ +/, $relstr);

    undef %temprels;
    foreach my $rf (@preds){
	my ($r, $fid) = split(/\|/, $rf);
	$temprels{$r}= $fid;
    }

    my $code = sub{ my $n = shift; return $n->height() == 0;};
    my @terminals = $node->find($code);
    #print $outfh $#terminals, "\n";

    $index = 1;
    foreach my $terminal (@terminals){
	if(!isTrace($terminal) || $keep_trace){
	    my $p = $terminal;

	    #go up to the level where it's no longer the head
	    while(defined($p) && isHead($p)){ 
		$p=$p->parent();
	    }

	    #search left siblings for head
	    my $lexhead = "DUMMY";
	    my $left = $p->prevSibling();
	
	    while(defined($left)){
		if(isHead($left)){
		    my $key = createID($left);
		    $lexhead = $heads{$key};
		}
		$left = $left->prevSibling();
	    }

	    #search right siblings for head
	    my $right = $p->nextSibling();
	
	    my $rightmost = $right;
	    while(defined($right)){
		if(isHead($right)){
		    my $key = createID($right);
		    $lexhead = $heads{$key};
		}
		if(!isTrace($right)){ #non-trace
		    $rightmost = $right;
		}
		$right = $right->nextSibling();
	    }

	    #if the current node is a conjunct and the immediate left sibling is a conjunction, 
	    #make that conjunction the head. This overwrites the normal algorithm of finding the head
	    if(isConjunct($p)&&
	       defined($p->prevSibling())&& 
	       isConjunction($p->prevSibling())){
		my $key = createID($p->prevSibling());
		$lexhead = $heads{$key};
	    }


	    #functional tags for node p
	    my @ftags = $p->ftags();
	    my $ftag = $ftags[0];

	    #index on for node p
	    my $traceid = $p->traceId();
	    if ($traceid){
		$traceid = "-" . $traceid;
		print $traceid;
	    }

	    #in case there is no head, pick the rightmost sibling as head
	    if($lexhead eq "DUMMY"){
		if(defined($rightmost)){
		    my $key = createID($rightmost);
		    $lexhead = $heads{$key};
		    if(!$ftag){
			$ftag = "UNK";
		    }
		}elsif(!$p->parent()){
		    #$p is rightmost node itself, and it has no parent
		    $lexhead = "0";
		    $ftag = "ROOT";
		}else{
		    print "Warning: Headless element that's not root! ", $p->rawdata(), "\n";
		}

	    }elsif(!$ftag){
		$ftag = "UNK";
	    }

	    
	    my $modftag = modify_ftag($ftag); #standardize functional tag/depdency label

	    my $pos = $terminal->label();
	    if(!$pos){
		$pos = "-"; #empty categories do not have POS
	    }
	

	    print $outfh  $index, "\t", $terminal->data(), $traceid, "\t", $terminal->data(), "\t", $terminal->data(), "\t", $pos, "\t", $pos, "\t", "_", "\t", "_", "\t", $lexhead, "\t", $lexhead, "\t", $modftag, "\t", $modftag;

	    if($temprels{$terminal->termno()}){
		#printf $outfh "%8s\t%12s\t","Y", $temprels{$terminal->termno()};
		print $outfh "\t","Y", "\t", $temprels{$terminal->termno()};
	    }else{
		#printf $outfh "%8s\t%10s\t","_", "_";
		print $outfh "\t","_", "\t", "_";
	    }
	    
	    foreach my $rf (sort{$a<=>$b} (@preds)){
		my ($r, $fid) = split(/\|/, $rf);
		if($argrel{"$r|$index"}){
		    my $arglabel = $argrel{"$r|$index"};
		    my $marglabel=modify_arglabel($arglabel);
		    #printf $outfh "%7s\t",  $marglabel;
		    print $outfh "\t",  $marglabel;
		}else{
		    #printf $outfh "%7s\t", "_";
		    print $outfh "\t", "_";
		}
	    }

	    print $outfh "\n";
	    if(!isTrace($terminal)){
		$index++;
	    }
	}
	#$index++;
    }
}


sub modify_arglabel{
    #standardize argument label
    my $label = shift;
    $label =~ s/^ARGM\-//;
    $label =~ s/ARG(\d)\-PSR$/A$1/;
    $label =~ s/ARG(\d)\-CRD$/A$1/;
    $label =~ s/ARG(\d)\-PSE$/C\-A$1/;
    $label =~ s/ARG(\d)\-PRD$/C\-A$1/;
    $label =~ s/ARG(\d)\-[A-Z]+$/C\-A$1/;
    $label =~ s/^ARG(\d)/A$1/;
    $label =~ s/^C\-ARG(\d)/C\-A$1/;
    return $label;
}

sub modify_ftag{
    #standardize dependency label
    my $tag = shift;
    $tag =~ s/^CONJUNCTION1/CJTN/;
    $tag =~ s/^CONJUNCTION\d+/cCJTN/;
    $tag =~ s/^CONJUNCT\d*/CJT/;
    $tag =~ s/^RELATIVE/RELC/;
    $tag =~ s/^UNSPEC/UNK/;
    return $tag;
}

sub isHead{
    #test if a node is a head
    my $node = shift;
    my @ftags = $node->ftags();
    foreach my $tag (@ftags){
	#print $tag, "=func\n";
	if($tag eq "HEAD"||
	   $tag eq "CONJUNCT1"||
	   #$tag eq "CONJUNCTION1"||
	   #$tag eq "PRD"||
	   $tag eq "SUBORD"){
	    return 1;
	}
	#if($tag eq "CONJUNCT1" && isConjunctHead($node)){
	#    return 1;
	#}
    }
    return 0;
}


sub isConjunctHead($node){
    #auxiliary function to test if a conjunct is a head
    #a conjunct is a head if it is the first conjunct AND there are no conjunctions
    my $node = shift;
    #search left for conjunctions
    my $left = $node->prevSibling();
    while(defined($left)){
	my @ftags = $left->ftags();
	foreach my $tag (@ftags){
	    if($tag =~ /CONJUNCTION/){
		return 0;
	    }
	}
	$left = $left->prevSibling();
    }
    #search right siblings for conjunctions
    my $right = $node->nextSibling();
    while(defined($right)){
	my @ftags = $right->ftags();
	foreach my $tag (@ftags){
	    if($tag =~ /CONJUNCTION/){
		return 0;
	    }
	}
	$right = $right->nextSibling();
    }
    return 1;
}



sub isLeaf{
    my ($node) = @_;
    if($node->height() == 0){
	return 1;
    }
    return 0;
}

sub isConj{
    my ($node) = @_;
    if($node->height() == 0 && 
       ($node->label() eq "CC" || 
	$node->label() eq "ETC" || 
	($node->label() eq "PU" &&
	 $node->data() ne ""))){
	return 1;
    }
    return 0;
}

sub isConjunction{
    my ($node) = @_;
    if($node->height() == 0){ 
 	my @ftags = $node->ftags();
	foreach my $tag (@ftags){
	    if($tag =~ /CONJUNCTION/){
		return 1;
	    }
	}
    }
    return 0;
}

sub isConjunct{
    my ($node) = @_;
    my @ftags = $node->ftags();
    foreach my $tag (@ftags){
	if($tag =~ /^CONJUNCT\d+$/){
	    return 1;
	}
    }
    return 0;
}

sub isCompound{
    my ($node) = @_;
    if($node->label() eq "VCD" || 
       $node->label() eq "VRD" || 
       $node->label() eq "VNV" || 
       $node->label() eq "VSB" || 
       $node->label() eq "VPT" || 
       $node->label() eq "VCP"){
	return 1;
    }
    return 0;
}

sub isFlat{
    my (@nodes) = @_;
    my $isflat = 1;
    foreach my $node (@nodes){
	if($node->height() != 0){
	    $isflat = 0;
	}
    }
    return $isflat;
}



sub isVerb{
    my ($node) = @_;
    if($node->label() =~ /^V/){
	return 1;
    }
    return 0;
}

sub isTrace{
    my ($node) = @_;
    if($node->rawLabel() =~ /NONE/i 
       || $node->data() =~ /\*[A-Za-z]+\*/
       || $node->data()=~/\*\-\d+/){
	return 1;
    }
    return 0;
}



sub printnode{
    my ($cnode, $outfh) = @_;
    print $outfh $cnode -> rawdata();
}


sub printConj{
    my ($outfh) = @_;
    print $outfh "(CONJOINED T)";
}


sub getFtags{
    my ($node) = @_;
    #put the nodes in a hash
    my %ftags;
    my @ftags = $node->ftags();
    foreach my $ftag (@ftags){
	$ftags{$ftag}++;
    }
    return %ftags;
}

sub printSemFeature{
    my ($outfh, $node) = @_;
    my @ftags = $node->ftags();
    my $ftag = $ftags[0];
    if(defined($ftag)){
	if($ftag eq "MNR"){
	    print $outfh "(SEM-FEATURE MNR)";
	}elsif($ftag eq "CND"){
	    print $outfh "(SEM-FEATURE CONDITION)";
	}elsif($ftag eq "DIR"){
	    print $outfh "(SEM-FEATURE DIR)";
	}elsif($ftag eq "PRP"){
	    print $outfh "(SEM-FEATURE PRP)";
	}elsif($ftag eq "EXT"){
	    print $outfh "(SEM-FEATURE EXT)";
	}elsif($ftag eq "Q"){
	    print $outfh "(MOOD INTERROGATIVE)";
	}elsif($ftag eq "FOC"){
	    print $outfh "(FOCUS T)";
	}elsif($ftag eq "HLN"){
	    print $outfh "(GENRE-TYPE HLN)";
	}elsif($ftag eq "TMP"){
	    print $outfh "(SEM-FEATURE TMP)";
	}elsif($ftag eq "IMP"){
	    print $outfh "(MOOD IMPERATIVE)";
	}elsif($ftag eq "TTL"){
	    print $outfh "(GENRE-TYPE TTL)";
	}elsif($ftag eq "LOC"){
	    print $outfh "(SEMFEATURE LOC)";
	}
    }

}

sub createID{
    my ($node) = @_;
    my $termno = $node -> termno();
    my $height = $node -> height();
    my $id = $termno . ":" . $height;
    return $id;
}


sub parsepb{
    local(*predargs, $pbfile)= @_;
    print $pbfile, "\n";
    open(PB, $pbfile);
    while(<PB>){
	my @ele = split;
	my $key = join(" ", @ele[0..2]);
	my $val = join(" ", @ele[6..$#ele]);
	$predargs{$key} = $val;
	#print $key, "\t", $val, "\n";
    }
    close(PB);
}

sub printpb{
    my ($args, $outfh) = @_;
    my @args = split(/\s+/, $args);
    foreach my $arg (@args){
	if(!($arg =~ /rel/i)){
	    my ($addr, $lbl) = split(/-/, $arg);
	    print $outfh "(P-$lbl ($labels{$addr} (EC-TYPE PB) (INDEX $indices{$addr})))";
	}
    }
}
